@@ -34,7 +34,7 @@
 #include <errno.h>
 #include "common.h"
 #include "lib.h"
-
+#include "dns.h"
 
 #ifdef DEBUG
 #define OPT_DEBUG 1
@@ -46,14 +46,19 @@
 /*
  * These are all the global variables.
  */
-int                 opt_debug = OPT_DEBUG;
+unsigned char
+       opt_debug = OPT_DEBUG;
 int                 opt_serv = 0;
 const char*         progname = 0;
+
+#ifdef ENABLE_PIDFILE
 #if defined(__sun__)
 const char*         pid_file = "/var/tmp/dnrd.pid";
 #else
 const char*         pid_file = "/var/run/dnrd.pid";
 #endif
+#endif
+
 int                 isock = -1;
 #ifdef ENABLE_TCP
 int                 tcpsock = -1;
@@ -61,23 +66,41 @@ int                 tcpsock = -1;
 int                 select_timeout = SELECT_TIMEOUT;
 int                 forward_timeout = FORWARD_TIMEOUT;
 //int                 load_balance = 0;
+#ifndef __CYGWIN__
 uid_t               daemonuid = 0;
 gid_t               daemongid = 0;
+char                dnrd_user[256] = "dnrd";
+char                dnrd_group[256] = "dnrd";
+#endif
 const char*         version = PACKAGE_VERSION;
 int                 gotterminal = 1; /* 1 if attached to a terminal */
 sem_t               dnrd_sem;  /* Used for all thread synchronization */
 
 int                 reactivate_interval = REACTIVATE_INTERVAL;
+int                 stats_interval = 0;
+int                 stats_reset = 1;
 
 /* The path where we chroot. All config files are relative this path */
-char chroot_path[512] = CHROOT_PATH;
+char                dnrd_root[512] = DNRD_ROOT;
 
-char config_file[] = CHROOT_PATH "/" CONFIG_FILE;
+char                config_file[512] = DNRD_ROOT "/" CONFIG_FILE;
 
-domnode_t *domain_list;
+domnode_t           *domain_list;
 /* turn this on to skip cache hits from responses of inactive dns servers */
 int                 ignore_inactive_cache_hits = 0; 
 
+/* highest socket number */
+int                 maxsock;
+
+/* maximum number of open sockets. If we have this amount of
+   concurrent queries, we start dropping new ones */
+int max_sockets = 200;
+
+/* the fd set. query modifies this so we make it global */
+fd_set              fdmaster;
+
+
+
 /*
  * This is the address we listen on.  It gets initialized to INADDR_ANY,
  * which means we listen on all local addresses.  Both the address and
@@ -94,6 +117,7 @@ struct sockaddr_in recv_addr = { AF_INET, 53, { INADDR_ANY } };
 /* init recv_addr in main.c instead of here */ 
 struct sockaddr_in recv_addr;
 
+#ifdef ENABLE_PIDFILE
 /* check if a pid is running 
  * from the unix faq
  * http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC18
@@ -150,9 +174,7 @@ int kill_current()
 	exit(-1);
     }
     if ((retn = (fscanf(filep, "%i%*s", &pid) == 1))) {
-	if (kill(pid, SIGTERM)) {
-	    log_msg(LOG_ERR, "Couldn't kill dnrd: %s", strerror(errno));
-	}
+        kill(pid, SIGTERM);
 	/* dnrd gets 4 seconds to die or we give up */
 	if (!wait_for_exit(pid, 400)) {
 	  log_msg(LOG_ERR, "The dnrd process didn't die within 4 seconds");
@@ -162,6 +184,32 @@ int kill_current()
     unlink(pid_file);
     return retn;
 }
+#endif /* ENABLE_PIDFILE*/
+
+const char *get_typestr(int type) {
+	static const char *EMERG = "EMERG: "; 
+	static const char *ALERT = "ALERT: "; 
+	static const char *CRIT = "CRIT:  "; 
+	static const char *ERR = "ERROR: "; 
+	static const char *WARNING = "Warning: "; 
+	static const char *NOTICE = "Notice: "; 
+	static const char *INFO = "Info:  "; 
+	static const char *DEBUG = "Debug: "; 
+	static const char *EMPTY = "";
+
+	switch (type) {
+	  case LOG_EMERG:   return EMERG;
+	  case LOG_ALERT:   return ALERT;
+	  case LOG_CRIT:    return CRIT;
+	  case LOG_ERR:     return ERR;
+	  case LOG_WARNING: return WARNING;
+	  case LOG_NOTICE:  return NOTICE;
+	  case LOG_INFO:    return INFO;
+	  case LOG_DEBUG:   return DEBUG;
+	default:          return EMPTY;
+	}
+}
+
 
 /*
  * log_msg()
@@ -180,24 +228,12 @@ void log_msg(int type, const char *fmt, ...)
     va_start(ap, fmt);
 
     if (gotterminal) {
-	const char *typestr;
-	switch (type) {
-	  case LOG_EMERG:   typestr = "EMERG: "; break;
-	  case LOG_ALERT:   typestr = "ALERT: "; break;
-	  case LOG_CRIT:    typestr = "CRIT:  "; break;
-	  case LOG_ERR:     typestr = "ERROR: "; break;
-	  case LOG_WARNING: typestr = "Warning: "; break;
-	  case LOG_NOTICE:  typestr = "Notice: "; break;
-	  case LOG_INFO:    typestr = "Info:  "; break;
-	  case LOG_DEBUG:   typestr = "Debug: "; break;
-	  default:          typestr = ""; break;
-	}
-	fprintf(stderr, typestr);
-	vfprintf(stderr, fmt, ap);
-	if (fmt[strlen(fmt) - 1] != '\n') fprintf(stderr, "\n");
+			fprintf(stderr, get_typestr(type));
+			vfprintf(stderr, fmt, ap);
+			if (fmt[strlen(fmt) - 1] != '\n') fprintf(stderr, "\n");
     }
     else {
-	vsyslog(type, fmt, ap);
+			vsyslog(type, fmt, ap);
     }
     va_end(ap);
 }
@@ -211,11 +247,11 @@ void log_msg(int type, const char *fmt, ...)
  * Abstract: If debugging is turned on, this will send the message
  *           to syslog with LOG_DEBUG priority.
  */
-void log_debug(const char *fmt, ...)
+void log_debug(int level, const char *fmt, ...)
 {
     va_list ap;
     
-    if (!opt_debug) return;
+    if (opt_debug < level) return;
 
     va_start(ap, fmt);
     if (gotterminal) {
@@ -244,7 +280,7 @@ void cleanexit(int status)
     /* Only let one process run this code) */
     sem_wait(&dnrd_sem);
 
-    log_debug("Shutting down...\n");
+    log_debug(1, "Shutting down...\n");
     if (isock >= 0) close(isock);
 #ifdef ENABLE_TCP
     if (tcpsock >= 0) close(tcpsock);
@@ -258,6 +294,35 @@ void cleanexit(int status)
     exit(status);
 }
 
+
+/*
+ * log_err_exit()
+ *
+ * In:      exitcode - the exitcode returned
+ *          fmt  - a formatting string, ala printf.
+ *          ...  - other printf-style arguments.
+ *
+ * Sends a message to log_msg, LOG_ERR and exit clean
+ */
+void log_err_exit(int exitcode, const char *fmt, ...)
+{
+    va_list ap;
+    va_start(ap, fmt);
+
+    if (gotterminal) {
+			fprintf(stderr, get_typestr(LOG_ERR));
+			vfprintf(stderr, fmt, ap);
+			if (fmt[strlen(fmt) - 1] != '\n') fprintf(stderr, "\n");
+    }
+    else {
+			vsyslog(LOG_ERR, fmt, ap);
+    }
+    va_end(ap);
+		cleanexit(exitcode);
+}
+
+
+
 /*
  * make_cname()
  *
@@ -300,60 +365,16 @@ char* make_cname(const char *text, const int maxlen)
     return cname;
 }
 
-
-
-void sprintf_cname(const char *cname, int namesize, char *buf, int bufsize)
-{
-  const char *s = cname; /*source pointer */
-  char *d = buf; /* destination pointer */
-
-  if (cname == NULL) return;
-    
-  if ((strnlen(cname, namesize)+1) > (unsigned)bufsize) {
-    if (bufsize > 11) {
-      sprintf(buf, "(too long)");
-    }
-    else {
-      buf[0] = 0;
-    }
-    return;
-  }
-
-  /* extract the pascal style strings */
-  while (*s) {
-    int i;
-    int size = *s;
-
-    /* Let us see if we are bypassing end of buffer.  Also remember
-     * that we need space for an ending \0
-     */
-    if ((s + *s - cname) >= (bufsize)) {
-      if (bufsize > 15 ) {
-	sprintf(buf, "(malformatted)");
-      } else {
-	buf[0] = 0;
-      }
-      return;
-    }
-
-    /* delimit the labels with . */
-    if (s++ != cname) sprintf(d++, ".");
-   
-    for(i = 0; i < size; i++) {
-      *d++ = *s++;
-    }
-    *d=0;
-  }
-}
-
 /* convert cname to ascii and return a static buffer */
+/* this func must *never* be called with an incomming DNS
+   packet as input. Never. */
 char *cname2asc(const char *cname) {
   static char buf[256];
   /* Note: we don't really check the size of the incomming cname. but
      according to RFC 1035 a name must not be bigger than 255 octets.
    */
   if (cname) 
-    sprintf_cname(cname, sizeof(buf), buf, sizeof(buf));
+    snprintf_cname((char *)cname, strlen(cname), 0, buf, sizeof(buf));
   else
     strncpy(buf, "(default)", sizeof(buf));
   return buf;
