@@ -33,8 +33,8 @@
 #include <string.h>
 #include <ctype.h>
 
-#include <signal.h>
 #include <netdb.h>
+#include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -42,15 +42,11 @@
 #include "common.h"
 #include "dns.h"
 #include "lib.h"
-
-#ifndef EXCLUDE_MASTER
 #include "master.h"
-#define	MASTER_CONFIG		CHROOT_PATH "/master"
-#endif
 
+#define	MASTER_CONFIG		"master"
+#define MASTER_BLACKLIST        "blacklist"
 #define	PACKET_ASSEMBLYSIZE	600
-
-
 #define	ARPADOMAIN		".in-addr.arpa"
 
 #define	DNS_TYPE_A		1
@@ -58,10 +54,8 @@
 #define	DNS_TYPE_PTR		12
 #define	DNS_TYPE_MX		15
 
-
 #define	DEFAULT_TTL		(60 * 60)
 
-
 typedef struct _string {
     unsigned int code;
     char	*string;
@@ -96,14 +90,14 @@ typedef struct _dnsrec {
     } u;
 } dnsrec_t;
 
-char master_param[200]		= "";
+unsigned char master_reload	= 0;
+int master_onoff		= 1;
+char master_config[256]		= MASTER_CONFIG;
+char blacklist[256]             = MASTER_BLACKLIST;
 
-static int master_onoff		= 1;
 static int master_initialised	= 0;
-static char config[200]		= MASTER_CONFIG;
 
 static int auto_authority	= 1;
-static int master_reload	= 0;
 
 	/*
 	 * The DNS database is stored in an array of dbmax length.  The
@@ -144,10 +138,13 @@ static int create_nameip(nameip_t *nameip, const int maxnamelen, char *ip)
 
     nameip->ipnum = ntohl(ipnum.s_addr);
     // Swap the bytes.  Don't assume 32-bit integers.
-    ipnum.s_addr = ((nameip->ipnum & 0xff000000 >> 24) |
-		    (nameip->ipnum & 0x00ff0000 >> 8) |
-		    (nameip->ipnum & 0x0000ff00 << 8) |
-		    (nameip->ipnum & 0x000000ff << 24));
+    if (nameip->ipnum == ipnum.s_addr)
+        ipnum.s_addr = (((nameip->ipnum & 0xff000000) >> 24) |
+		        ((nameip->ipnum & 0x00ff0000) >> 8 ) |
+		        ((nameip->ipnum & 0x0000ff00) << 8 ) |
+		        ((nameip->ipnum & 0x000000ff) << 24));
+    else
+        ipnum.s_addr = nameip->ipnum;
 
     mkstring(&nameip->arpa, inet_ntoa(ipnum), maxnamelen);
     
@@ -213,6 +210,7 @@ static int reset_master(void)
     if (dbv)
     {
 	free(dbv);
+	dbv = NULL;
 	dbc = 0;
     }
 
@@ -273,7 +271,7 @@ static dnsrec_t *add_authority(char *domain, const int maxlen)
     rec = alloc_dnsrec(domain, maxlen);
     rec->type = DNS_AUTHORITY;
     add_record(rec);
-    log_debug("added authority for %s", domain);
+    log_debug(1, "added authority for %s", domain);
 
     return (rec);
 }
@@ -330,10 +328,13 @@ int read_hosts(char *filename, char *domain)
     FILE	*fp;
     
     if ((fp = fopen(filename, "r")) == NULL) {
-	log_msg(LOG_NOTICE, "can't open file: %s", filename);
+	log_msg(LOG_ERR, "can't open file: %s", filename);
 	return (1);
     }
 
+    log_debug(1, "Reading hosts from %s/%s, domain= %s", dnrd_root,
+	      filename, *domain == 0? "<none>": domain);
+
     count = dbc;
     while (fgets(line, sizeof(line), fp) != NULL) {
 	p = skip_ws((char *)noctrln(line, sizeof(line)));
@@ -353,7 +354,7 @@ int read_hosts(char *filename, char *domain)
     }
 	    
     fclose (fp);
-    log_debug("%s: %d records", filename, dbc - count);
+    log_debug(1, "%s: %d records", filename, dbc - count);
 
     return (0);
 }
@@ -362,11 +363,11 @@ int read_configuration(char *filename)
 {
     int	count;
     char *p, word[100], ipnum[100], line[300];
-    char domain[40];
+    char domain[256];
     FILE *fp;
     
     if ((fp = fopen(filename, "r")) == NULL) {
-	log_debug("no master configuration: %s", filename);
+	log_debug(1, "no master configuration: %s", filename);
 	return (1);
     }
 
@@ -382,9 +383,16 @@ int read_configuration(char *filename)
 	     * name makes one DNS record.
 	     */
 	    get_word(&p, ipnum, sizeof(ipnum));
+	    if (*get_hostname(&p, domain, word, sizeof(word)) == 0) {
+	      /* no host name specified. Lets assign the domainname the ip*/
+	      add_nameip(domain, sizeof(domain), ipnum);
+	    } else {
+	      /* add the hostnames */
+	      add_nameip(word, sizeof(word), ipnum);
 	    while (*get_hostname(&p, domain, word, sizeof(word)) != 0)
 		    add_nameip(word, sizeof(word), ipnum);
 	}
+	}
 	else if (strcmp(get_word(&p, word, sizeof(word)), "domain") == 0) {
 	    get_word(&p, domain, sizeof(domain));
 	    if (*domain == 0) continue;
@@ -419,7 +427,7 @@ int read_configuration(char *filename)
     }
 	    
     fclose (fp);
-    log_debug("%s: %d records", filename, dbc - count);
+    log_debug(1, "%s: %d records", filename, dbc - count);
 
     return (0);
 }
@@ -690,80 +698,74 @@ static int compile_dnsrecords(dnsheader_t *x, char *object)
 }
 
 
+/* 
+ * read_blaclist()
+ *
+ * Reads the blacklist and adds authorities for all the blacklisted hosts.
+ */
+int read_blacklist(const char *filename) {
+  FILE *f;
+  static char buf[256];
+
+  if ( (f=fopen(filename, "r")) == NULL) {
+    log_debug(1, "no blacklist: %s", filename);
+    return (1);
+  }
+  log_msg(LOG_NOTICE, "Reading blacklist %s", filename);
+  while (fgets(buf, sizeof(buf), f) != NULL) {
+    char *p=strchr(buf, '#');
+    if (p != NULL) *p='\0'; /* ignore comments */
+    if ((p = strchr(buf, '\r')) != NULL) *p='\0'; /* so we can read dos files*/
+    if ((p = strchr(buf, '\n')) != NULL) *p='\0';
+    if (buf[0]) add_authority(buf, sizeof(buf));
+  } 
+  return (0);
+}
+
+
 /*
  * master_init()
  *
  * Create the DNS database with the data from the configuration
  * file.
  */
-static int _master_init(void)
+int master_init(void)
 {
-    if (master_onoff == 0) return (0);
-
-    log_debug("initialising master DNS database");
+    if (master_onoff == 0) {
+	log_msg(LOG_NOTICE, "local DNS master turned off");
+	return (0);
+    }
+    log_debug(1, "initialising master DNS database");
 
     add_nameip("localhost", sizeof("localhost"), "127.0.0.1");
     add_dns("0.0.127.in-addr.arpa", sizeof("0.0.127.in-addr.arpa"),
 	    "localhost", sizeof("localhost"));
     
-    if ((strcmp(master_param, "hosts") == 0) ||
-	(read_configuration(config) != 0)) {
-	char	domain[80];
-	FILE	*fp;
-
-	/*
-	 * Hmm, no dnrd.conf - let's try to initialize with the
-	 * /etc/hosts only.
-	 */
-
-	*domain = 0;
-	if ((fp = fopen("/etc/resolv.conf", "r")) != NULL) {
-	    char	*p, word[80], line[300];
-
-	    /*
-	     * We try to get the default domain name from here.
-	     */
-
-	    while (fgets(line, sizeof(line), fp) != NULL) {
-		p = skip_ws(noctrln(line,sizeof(line)));
-		if (*p == 0  ||  *p == '#') continue;
-
-		get_word(&p, word, sizeof(word));
-		if (strcmp(word, "domain") == 0) {
-		    get_word(&p, domain, sizeof(domain));
-		    break;
-		}
-	    }
-
-	    fclose (fp);
-
-	    log_debug("initialising from /etc/hosts, domain= %s",
-		      *domain == 0? "<none>": domain);
-	    read_hosts("/etc/hosts", domain);
-	}
+    read_blacklist(blacklist);
+    if (read_configuration(master_config) != 0) {
+		/* falied to read coifg */
     }
 
     if (auto_authority != 0) {
-	int	i;
-	char	*domain, arpaname[200];
+		int	i;
+		char	*domain, arpaname[200];
 	
-	
-	for (i = 0; i < dbc; i++) {
-	    if (dbv[i]->type == DNS_NAMEIP) {
-		snprintf (arpaname, sizeof(arpaname) - 2, "%s%s",
-			  dbv[i]->u.nameip.arpa.string, ARPADOMAIN);
-		if ((domain = strchr(arpaname, '.')) == NULL) continue;
-		domain++;
-		if (authority_lookup(domain) == 0) {
-		    add_authority(domain,sizeof(arpaname)-(arpaname-domain));
+		for (i = 0; i < dbc; i++) {
+			if (dbv[i]->type == DNS_NAMEIP) {
+				snprintf (arpaname, sizeof(arpaname) - 2, "%s%s",
+						  dbv[i]->u.nameip.arpa.string, ARPADOMAIN);
+				if ((domain = strchr(arpaname, '.')) == NULL) continue;
+				domain++;
+				if (authority_lookup(domain) == 0) {
+					add_authority(domain,sizeof(arpaname)-(arpaname-domain));
+				}
+			}
 		}
-	    }
-	}
     }
-
-    log_debug("%d records in master DNS database", dbc);
+	
+    log_debug(1, "%d records in master DNS database", dbc);
     master_initialised = 1;
-
+	
     return (0);
 }
 
@@ -794,10 +796,10 @@ int master_lookup(unsigned char *msg, int len)
 
 
     if (master_initialised == 0) {
-	_master_init();
+	master_init();
     }
 
-    if ((parse_query(&query, msg, len) == NULL) ||
+    if (parse_query(&query, msg, len) ||
 	(query.class != DNS_CLASS_INET  ||  GET_OPCODE(query.flags) != 0)) {
 	return (0);
     }
@@ -816,7 +818,7 @@ int master_lookup(unsigned char *msg, int len)
 	    dnsheader_t *x;
 	    
 	    query.name[k] = '.';
-	    log_debug("master: found PTR %s\n", query.name);
+	    log_debug(2, "master: found PTR %s\n", query.name);
 
 	    x = begin_assembly(&query);
 	    compile_objectname(x);
@@ -917,10 +919,11 @@ int master_lookup(unsigned char *msg, int len)
     }
 
     domain++;
-    if (authority_lookup(domain) != NULL) {
+    if ((authority_lookup(query.name) != NULL) 
+	||  (authority_lookup(domain) != NULL)) {
 	dnsheader_t *x;
 
-	log_debug("master: found AUTHORITY for %s\n", domain);
+	log_debug(2, "master: found AUTHORITY for %s\n", query.name);
 
 	x = begin_assembly(&query);
 	x->ancount = 0;
@@ -954,10 +957,10 @@ int master_dontknow(unsigned char *msg, int len, unsigned char *answer)
 
 
     if (master_initialised == 0) {
-	_master_init();
+	master_init();
     }
 
-    if (parse_query(&query, msg, len) == NULL) {
+    if (parse_query(&query, msg, len)) {
 	return (0);
     }
     else if (query.class != DNS_CLASS_INET  ||  GET_OPCODE(query.flags) != 0) {
@@ -976,33 +979,25 @@ int master_dontknow(unsigned char *msg, int len, unsigned char *answer)
     return (x->len);
 }
 
-
 /*
- * master_sighup(), master_reinit()
+ * master_reinit()
  *
  * master_reinit() erases the whole master DNS database and makes a
- * complete reread of the database definition.  master_sighup()
- * is the correponding signal handler that sets the reload flag.
- * The actual reload is done from run() in relay.c after return
- * from the select() function.
+ * complete reread of the database definition. The actual reload is
+ * done from run() in relay.c after return from the pselect()
+ * function.
  */
-static void master_sighup(int sig)
-{
-    if (master_onoff != 0) {
-        master_reload = 1;
-    }
-
-    signal(sig, master_sighup);
-    return;
-}
 
 int master_reinit(void)
 {
-    if (master_onoff == 0) return (0);
+    if (master_onoff == 0) {
+      master_reload = 0;
+      return (0);
+    }
 
     if (master_reload != 0) {
         reset_master();
-        _master_init();
+        master_init();
     }
 
     master_reload = 0;
@@ -1010,32 +1005,4 @@ int master_reinit(void)
 }
 
 
-/*
- * master_init()
- *
- * Should be called once from main() to initialise the master
- * database.
- */ 
-int master_init(void)
-{
-    if (strcmp(master_param, "off") == 0) master_onoff = 0;
-
-    if (master_onoff == 0) {
-	log_msg(LOG_NOTICE, "local DNS master turned off");
-    }
-    else {
-	_master_init();
-	signal(SIGHUP, master_sighup);
-    }
-
-    /*
-     * Since our root dir will have changed to /etc/dnrd, we need to change
-     * the path of our config file so that future re-reads will continue to
-     * work.
-     */
-    strcpy(config, "master");
-
-    return (0);
-}
-
 #endif
