@@ -23,7 +23,8 @@ int myrand(int max) {
 
 void qid_init_pool(void) {
   int i;
-  FILE *f;
+  /*  FILE *f;
+   */
   struct timeval tv;
 
   for (i=0; i<QID_POOL_SIZE; i++)
@@ -40,7 +41,7 @@ void qid_init_pool(void) {
 
 unsigned short int qid_get(void) {
   unsigned short int t;
-  int i = myrand(pool_ptr+1);
+  int i = myrand(pool_ptr) % QID_POOL_SIZE;
   /*
   if (pool_ptr == 0)
   log_debug("return_qid: qid pool is empty.");*/
