@@ -26,12 +26,17 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <syslog.h>
 #include <string.h>
+#include <time.h>
 #include <assert.h>
 #include "common.h"
 #include "relay.h"
 #include "cache.h"
 #include "query.h"
+#include "domnode.h"
+#include "check.h"
+#include "dns.h"
 
 #ifndef EXCLUDE_MASTER
 #include "master.h"
@@ -45,31 +50,33 @@
  *
  * Returns:  The return code from sendto().
  */
-static int dnssend(srvnode_t *s, void *msg, int len)
+static int udp_send(int sock, srvnode_t *srv, void *msg, int len)
 {
     int	rc;
     time_t now = time(NULL);
-
-    rc = sendto(s->sock, msg, len, 0,
-		(const struct sockaddr *) &s->addr,
+    rc = sendto(sock, msg, len, 0,
+		(const struct sockaddr *) &srv->addr,
 		sizeof(struct sockaddr_in));
 
     if (rc != len) {
-	log_msg(LOG_ERR, "sendto error: %s: %m",
-		inet_ntoa(s->addr.sin_addr));
+	log_msg(LOG_ERR, "sendto error: %s: ",
+		inet_ntoa(srv->addr.sin_addr), strerror(errno));
 	return (rc);
     }
-    if ((s->send_time == 0)) s->send_time = now;
+    if ((srv->send_time == 0)) srv->send_time = now;
+    srv->send_count++;
     return (rc);
 }
 
-/* send a query to current server. Returns len on success and 0 on
-   failure */
-
-int send2current(domnode_t *d, void *msg, const int len) {
+int send2current(query_t *q, void *msg, const int len) {
     /* If we have domains associated with our servers, send it to the
        appropriate server as determined by srvr */
-  while ((d->current != NULL) && (dnssend(d->current, msg, len) != len)) {
+  domnode_t *d;
+  assert(q != NULL);
+  assert(q->domain != NULL);
+
+  d = q->domain;
+  while ((d->current != NULL) && (udp_send(q->sock, d->current, msg, len) != len)) {
     if (reactivate_interval) deactivate_current(d);
   }
   if (d->current != NULL) {
@@ -77,6 +84,7 @@ int send2current(domnode_t *d, void *msg, const int len) {
   } else return 0;
 }
 
+
 /*
  * handle_udprequest()
  *
@@ -84,46 +92,63 @@ int send2current(domnode_t *d, void *msg, const int len) {
  * know the correct reply via master, caching, etc.), or forwarding them to
  * an appropriate DNS server.
  */
-void handle_udprequest()
+query_t *udp_handle_request()
 {
     unsigned           addr_len;
     int                len;
-    const int          maxsize = 512; /* According to RFC 1035 */
-    char               msg[maxsize+4];/* Do we really want this on the stack?*/
+    const int          maxsize = UDP_MAXSIZE;
+    static char        msg[UDP_MAXSIZE+4];
     struct sockaddr_in from_addr;
     int                fwd;
     domnode_t          *dptr;
+    query_t *q, *prev;
 
     /* Read in the message */
     addr_len = sizeof(struct sockaddr_in);
     len = recvfrom(isock, msg, maxsize, 0,
 		   (struct sockaddr *)&from_addr, &addr_len);
     if (len < 0) {
-	log_debug("recvfrom error %s", strerror(errno));
-	return;
-    }
-    if (len > maxsize) {
-	log_msg(LOG_WARNING, "Received message is too big to process");
-	return;
-    }
-    if (len < 12) {
-      return; /* too small */
+	log_debug(1, "recvfrom error %s", strerror(errno));
+	return NULL;
     }
+
+    /* do some basic checking */
+    if (check_query(msg, len) < 0) return NULL;
+
     /* Determine how query should be handled */
     if ((fwd = handle_query(&from_addr, msg, &len, &dptr)) < 0)
-      return; /* if its bogus, just ignore it */
+      return NULL; /* if its bogus, just ignore it */
 
     /* If we already know the answer, send it and we're done */
     if (fwd == 0) {
 	if (sendto(isock, msg, len, 0, (const struct sockaddr *)&from_addr,
 		   addr_len) != len) {
-	    log_debug("sendto error %s", strerror(errno));
+	    log_debug(1, "sendto error %s", strerror(errno));
 	}
-	return;
+	return NULL;
     }
 
-    dnsquery_add(&from_addr, msg, len);
-    if (!send2current(dptr, msg, len)) {
+
+    /* dptr->current should never be NULL it is checked in handle_query */
+
+    //    dnsquery_add(&from_addr, msg, len);
+    // if (!send2current(dptr, msg, len)) {
+
+    /* rewrite msg, get id and add to list*/
+
+    if ((prev=query_add(dptr, dptr->current, &from_addr, msg, len)) == NULL){
+      /* of some reason we could not get any new queries. we have to
+	 drop this packet */
+      return NULL;
+    }
+    q = prev->next;
+
+
+    if (send2current(q, msg, len) > 0) {
+      /* add to query list etc etc */
+      return q;
+    } else {
+
       /* we couldn't send the query */
 #ifndef EXCLUDE_MASTER
       int	packetlen;
@@ -142,18 +167,17 @@ void handle_udprequest()
        */
       
       if ((packetlen = master_dontknow(msg, len, packet)) > 0) {
-	   if (!dnsquery_find(msg, &from_addr)) {
-	   log_debug("ERROR: couldn't find the original query");
-	   return;
-	   }
+	query_delete_next(prev);
+	return NULL;
 	if (sendto(isock, msg, len, 0, (const struct sockaddr *)&from_addr,
 		   addr_len) != len) {
-	  log_debug("sendto error %s", strerror(errno));
-	  return;
+	  log_debug(1, "sendto error %s", strerror(errno));
+	  return NULL;
 	}
       }
 #endif
     }
+    return q;
 }
 
 /*
@@ -165,26 +189,26 @@ void handle_udprequest()
  * Returns:  A positove number indicating of the bytes received, -1 on a
  *           recvfrom error and 0 if the received message is too large.
  */
-static int dnsrecv(srvnode_t *s, void *msg, int len)
+static int reply_recv(query_t *q, void *msg, int len)
 {
     int	rc, fromlen;
     struct sockaddr_in from;
 
     fromlen = sizeof(struct sockaddr_in);
-    rc = recvfrom(s->sock, msg, len, 0,
+    rc = recvfrom(q->sock, msg, len, 0,
 		  (struct sockaddr *) &from, &fromlen);
 
     if (rc == -1) {
-	log_msg(LOG_ERR, "recvfrom error: %s: %m",
-		inet_ntoa(s->addr.sin_addr));
+	log_msg(LOG_ERR, "recvfrom error: %s",
+		inet_ntoa(q->srv->addr.sin_addr));
 	return (-1);
     }
     else if (rc > len) {
 	log_msg(LOG_NOTICE, "packet too large: %s",
-		inet_ntoa(s->addr.sin_addr));
+		inet_ntoa(q->srv->addr.sin_addr));
 	return (0);
     }
-    else if (memcmp(&from.sin_addr, &s->addr.sin_addr,
+    else if (memcmp(&from.sin_addr, &q->srv->addr.sin_addr,
 		    sizeof(from.sin_addr)) != 0) {
 	log_msg(LOG_WARNING, "unexpected server: %s",
 		inet_ntoa(from.sin_addr));
@@ -200,81 +224,109 @@ static int dnsrecv(srvnode_t *s, void *msg, int len)
  * This function handles udp DNS requests by either replying to them (if we
  * know the correct reply via master, caching, etc.), or forwarding them to
  * an appropriate DNS server.
+ *
+ * Note that the mached query is prev->next and not prev.
  */
-void handle_udpreply(srvnode_t *srv)
+void udp_handle_reply(query_t *prev)
 {
-    const int          maxsize = 512; /* According to RFC 1035 */
-    char               msg[maxsize+4];/* Do we really want this on the stack?*/
+  //    const int          maxsize = 512; /* According to RFC 1035 */
+    static char        msg[UDP_MAXSIZE+4];
     int                len;
-    struct sockaddr_in from_addr;
     unsigned           addr_len;
+    query_t *q = prev->next;
+
+    log_debug(3, "handling socket %i", q->sock);
+    if ((len = reply_recv(q, msg, UDP_MAXSIZE)) < 0)
+      {
+	log_debug(1, "dnsrecv failed: %i", len);
+	query_delete_next(prev);
+	return; /* recv error */
+      }
+    
+    /* do basic checking */
+    if (check_reply(q->srv, msg, len) < 0) {
+      log_debug(1, "check_reply failed");
+      query_delete_next(prev);
+      return;
+    }
 
-    len = dnsrecv(srv, msg, maxsize);
     if (opt_debug) {
 	char buf[256];
-	sprintf_cname(&msg[12], len-12, buf, 256);
-	log_debug("Received DNS reply for \"%s\"", buf);
+	snprintf_cname(msg, len, 12, buf, sizeof(buf));
+	log_debug(3, "Received DNS reply for \"%s\"", buf);
     }
-    if (len > 0) {
 
-      dump_dnspacket("reply", msg, len);
-      addr_len = sizeof(struct sockaddr_in);
-      if (!dnsquery_find(msg, &from_addr)) {
-	if (!srv->inactive) {
-	  log_debug("ERROR: couldn't find the original query");
-#ifdef DEBUG
-	  dnsquery_dump();
-#endif
-	}
+    dump_dnspacket("reply", msg, len);
+    addr_len = sizeof(struct sockaddr_in);
+
+    /* was this a dummy reactivate query? */
+    if (q->domain != NULL) {
+      /* no, lets cache the reply and send to client */
+      cache_dnspacket(msg, len, q->srv);
+
+      /* set the client qid */
+      *((unsigned short *)msg) = q->client_qid;
+      log_debug(3, "Forwarding the reply to the host %s",
+		inet_ntoa(q->client.sin_addr));
+      if (sendto(isock, msg, len, 0,
+		 (const struct sockaddr *)&q->client,
+		 addr_len) != len) {
+	log_debug(1, "sendto error %s", strerror(errno));
       }
-      else {
-	cache_dnspacket(msg, len, srv);
-	log_debug("Forwarding the reply to the host");
-	if (sendto(isock, msg, len, 0,
-		   (const struct sockaddr *)&from_addr,
-		   addr_len) != len) {
-	  log_debug("sendto error %s", strerror(errno));
-	}
-	
-      }
-      
-      /* this server is obviously alive, we reset the counters */
-      srv->send_count = 0; /* this is not used anymore? */
-      srv->send_time = 0;
-      if (srv->inactive) log_debug("Reactivating server %s",
-				   inet_ntoa(srv->addr.sin_addr));
-      srv->inactive = 0;
+    } else {
+      log_debug(2, "We got a reactivation dummy reply. Cool!");
     }
+      
+    /* this server is obviously alive, we reset the counters */
+    q->srv->send_time = 0;
+    if (q->srv->inactive) log_debug(1, "Reactivating server %s",
+				 inet_ntoa(q->srv->addr.sin_addr));
+    q->srv->inactive = 0;
+    /* remove query from list and destroy it */
+    query_delete_next(prev);
 }
 
 
 /* send a dummy packet to a deactivated server to check if its back*/
-int send_dummy(srvnode_t *s) {
+int udp_send_dummy(srvnode_t *s) {
   static unsigned char dnsbuf[] = {
   /* HEADER */
-
+    /* we send a lookup for localhost */
+    /* will this work on a big endian system? */
     0x00, 0x00, /* ID */
-    0x00, 0x00, /* QR|OC|AA|TC|RD -  RA|Z|RCODE  */
+    0x01, 0x00, /* QR|OC|AA|TC|RD -  RA|Z|RCODE  */
     0x00, 0x01, /* QDCOUNT */
     0x00, 0x00, /* ANCOUNT */
     0x00, 0x00, /* NSCOUNT */
     0x00, 0x00, /* ARCOUNT */
-    
-    /* QNAME */
-    9, 'l','o','c','a','l','h','o','s','t',0,
-    /* QTYPE */
-    0x00,0x01,   /* A record */
-    
-    /* QCLASS */
-    0x00,0x01   /* IN */
+
+    9, 'l','o','c','a','l','h','o','s','t',0,  /* QNAME */
+    0x00,0x01,   /* QTYPE A record */
+    0x00,0x01   /* QCLASS: IN */
+
+    /* in case you want to lookup root servers instead, use this: */
+    /*    0x00,       */ /* QNAME:  empty */
+    /*    0x00, 0x02, */ /* QTYPE:  a authorative name server */
+    /*    0x00, 0x01  */ /* QCLASS: IN */
   };
-  
-  /* send the packet */
+  query_t *q;
+  struct sockaddr_in srcaddr;
 
   /* should not happen */
   assert(s != NULL);
 
-  log_debug("Sending dummy id=%i to %s",  (unsigned short int) dnsbuf[0]++,
-	    inet_ntoa(s->addr.sin_addr));
-  return dnssend(s, &dnsbuf, sizeof(dnsbuf));
+  if ((q=query_add(NULL, s, &srcaddr, dnsbuf, sizeof(dnsbuf))) != NULL) {
+    int rc;
+    q = q->next; /* query add returned the query 1 before in list */
+    /* don't let those queries live too long */
+    q->ttl = reactivate_interval;
+    memset(&srcaddr, 0, sizeof(srcaddr));
+    log_debug(2, "Sending dummy id=%i to %s", ((unsigned short *)dnsbuf)[0], 
+	      inet_ntoa(s->addr.sin_addr));
+    /*  return dnssend(s, &dnsbuf, sizeof(dnsbuf)); */
+    rc=udp_send(q->sock, s, dnsbuf, sizeof(dnsbuf));
+    ((unsigned short *)dnsbuf)[0]++;
+    return rc;
+  }
+  return -1;
 }
