@@ -21,15 +21,17 @@
 #ifndef _DNRD_UDP_H_
 #define _DNRD_UDP_H_
 #include "srvnode.h"
+#include "query.h"
 
 /* Function to call when a message is available on isock */
-void handle_udprequest();
+/* returns the new query */
+query_t *udp_handle_request();
 
 /* Call this to handle upd DNS replies */
-void handle_udpreply(srvnode_t *srv);
+void udp_handle_reply(query_t *q);
 
 /* send a reactivation packet */
-int send_dummy(srvnode_t *s);
+int udp_send_dummy(srvnode_t *s);
 
 
 #endif /* _DNRD_UDP_H_ */
